/*===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
*/
#include <align/extern.h>

#include <klib/rc.h>
#include <insdc/insdc.h>
#include <vdb/database.h>
#include <align/writer-alignment.h>
#include "writer-priv.h"
#include "reader-cmn.h"
#include "debug.h"
#include <sysalloc.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <assert.h>

static const TableWriterColumn TableWriterAlgn_cols[ewalgn_cn_Last + 1] =
{
    /* order is important, see enum in .h !!! */
    {0, "TMP_KEY_ID", sizeof(uint32_t) * 8, ewcol_Temporary | ewcol_Ignore},
    {0, "PLOIDY", sizeof(uint32_t) * 8, ewcol_Ignore},
    {0, "SEQ_SPOT_ID", sizeof(int64_t) * 8, ewcol_IsArray | ewcol_Ignore},
    {0, "SEQ_READ_ID", sizeof(INSDC_coord_one) * 8, ewcol_IsArray },
    {0, "READ_START", sizeof(INSDC_coord_zero) * 8, ewcol_IsArray | ewcol_Ignore},
    {0, "READ_LEN", sizeof(INSDC_coord_len) * 8, ewcol_IsArray | ewcol_Ignore},
    {0, "REF_ID", sizeof(int64_t) * 8, ewcol_IsArray | ewcol_Ignore},
    {0, "REF_START", sizeof(INSDC_coord_zero) * 8, ewcol_IsArray | ewcol_Ignore},
    {0, "GLOBAL_REF_START", sizeof(uint64_t) * 8, ewcol_IsArray},
    {0, "REF_LEN", sizeof(INSDC_coord_len) * 8, ewcol_IsArray | ewcol_Ignore},
    {0, "REF_ORIENTATION", sizeof(bool) * 8, ewcol_IsArray},
    {0, "REF_PLOIDY", sizeof(uint32_t) * 8, ewcol_IsArray | ewcol_Ignore},
    {0, "MATE_REF_ORIENTATION", sizeof(bool) * 8, ewcol_Ignore},
    {0, "MATE_REF_ID", sizeof(int64_t) * 8, ewcol_Ignore},
    {0, "MATE_REF_POS", sizeof(INSDC_coord_zero) * 8, ewcol_Ignore},
    {0, "MATE_ALIGN_ID", sizeof(int64_t) * 8, ewcol_Ignore},
    {0, "TEMPLATE_LEN", sizeof(INSDC_coord_zero) * 8, ewcol_Ignore},
    {0, "MAPQ", sizeof(int32_t) * 8, ewcol_IsArray},
    {0, "HAS_MISMATCH", sizeof(bool) * 8, ewcol_IsArray},
    {0, "(bool)HAS_REF_OFFSET", sizeof(bool) * 8, ewcol_IsArray},
    {0, "MISMATCH", sizeof(INSDC_dna_text) * 8, ewcol_IsArray},
    {0, "REF_OFFSET", sizeof(int32_t) * 8, ewcol_IsArray}
};

static const TableReaderColumn TableAlgnReadTmpKey_cols[] = {
    {0, "TMP_KEY_ID", NULL, 0, 0},
    {0, NULL, NULL, 0, 0}
};

struct TableWriterAlgn {
    uint32_t options;
    const TableWriter* base;
    uint8_t cursor_id;
    TableWriterColumn cols[sizeof(TableWriterAlgn_cols)/sizeof(TableWriterAlgn_cols[0])];
    uint8_t spotid_cursor_id;
    TableWriterColumn cols_spotid[1];
    const TableReader* tmpkey_reader;
    TableReaderColumn cols_read_tmpkey[sizeof(TableAlgnReadTmpKey_cols) / sizeof(TableAlgnReadTmpKey_cols[0])];
};

LIB_EXPORT rc_t CC TableWriterAlgn_Make(const TableWriterAlgn** cself, VDatabase* db,
                                        ETableWriterAlgn_TableType type, uint32_t options)
{
    rc_t rc = 0;
    TableWriterAlgn* self = NULL;
    const char* tbl_nm = __func__;

    if( cself == NULL || db == NULL ) {
        rc = RC(rcAlign, rcFormatter, rcConstructing, rcParam, rcNull);
    } else {
        self = calloc(1, sizeof(*self));
        if( self == NULL ) {
            rc = RC(rcAlign, rcFormatter, rcConstructing, rcMemory, rcExhausted);
        } else {
            memcpy(self->cols, TableWriterAlgn_cols, sizeof(TableWriterAlgn_cols));
            switch(type) {
            case ewalgn_tabletype_PrimaryAlignment:
                tbl_nm = "PRIMARY_ALIGNMENT";
                break;
            case ewalgn_tabletype_SecondaryAlignment:
                tbl_nm = "SECONDARY_ALIGNMENT";
                self->cols[ewalgn_cn_REF_LEN].flags &= ~ewcol_Ignore;
                self->cols[ewalgn_cn_HAS_MISMATCH].flags |= ewcol_Ignore;
                self->cols[ewalgn_cn_MISMATCH].flags |= ewcol_Ignore;

                self->cols[ewalgn_cn_MATE_REF_ORIENTATION].flags &= ~ewcol_Ignore;
                self->cols[ewalgn_cn_MATE_REF_ID].flags &= ~ewcol_Ignore;
                self->cols[ewalgn_cn_MATE_REF_POS].flags &= ~ewcol_Ignore;
                self->cols[ewalgn_cn_MATE_ALIGN_ID].flags &= ~ewcol_Ignore;
                self->cols[ewalgn_cn_TEMPLATE_LEN].flags &= ~ewcol_Ignore;
                break;
            case ewalgn_tabletype_EvidenceInterval:
                tbl_nm = "EVIDENCE_INTERVAL";
                options |= ewalgn_co_unsorted;
                options |= ewalgn_co_PLOIDY;
                self->cols[ewalgn_cn_SEQ_READ_ID].flags |= ewcol_Ignore;
                break;
            case ewalgn_tabletype_EvidenceAlignment:
                tbl_nm = "EVIDENCE_ALIGNMENT";
                self->cols[ewalgn_cn_REF_LEN].flags &= ~ewcol_Ignore;
                self->cols[ewalgn_cn_REF_PLOIDY].flags &= ~ewcol_Ignore;
                self->cols[ewalgn_cn_HAS_MISMATCH].flags |= ewcol_Ignore;
                self->cols[ewalgn_cn_MISMATCH].flags |= ewcol_Ignore;
                options |= ewalgn_co_unsorted;
                break;
            default:
                rc = RC(rcAlign, rcFormatter, rcConstructing, rcType, rcUnrecognized);
            }
            if( options & ewalgn_co_SEQ_SPOT_ID ) {
                self->cols[ewalgn_cn_SEQ_SPOT_ID].flags &= ~ewcol_Ignore;
            }
            if( options & ewalgn_co_TMP_KEY_ID ) {
                self->cols[ewalgn_cn_TMP_KEY_ID].flags &= ~ewcol_Ignore;
            }
            if( options & ewalgn_co_PLOIDY ) {
                self->cols[ewalgn_cn_PLOIDY].flags &= ~ewcol_Ignore;
                self->cols[ewalgn_cn_READ_START].flags &= ~ewcol_Ignore;
                self->cols[ewalgn_cn_READ_LEN].flags &= ~ewcol_Ignore;
            }
            if( options & ewalgn_co_unsorted ) {
                self->cols[ewalgn_cn_REF_ID].flags &= ~ewcol_Ignore;
                self->cols[ewalgn_cn_REF_START].flags &= ~ewcol_Ignore;
                self->cols[ewalgn_cn_GLOBAL_REF_START].flags |= ewcol_Ignore;
            }
            if( (rc = TableWriter_Make(&self->base, db, tbl_nm)) == 0 ) {
                rc = TableWriter_AddCursor(self->base, self->cols, sizeof(self->cols)/sizeof(self->cols[0]), &self->cursor_id);
                self->options = options;
            }
        }
    }
    if( rc == 0 ) {
        *cself = self;
        ALIGN_DBG("table %s created", tbl_nm);
    } else {
        TableWriterAlgn_Whack(self, false, NULL);
        ALIGN_DBGERR(rc);
    }
    return rc;
}

LIB_EXPORT rc_t CC TableWriterAlgn_Whack(const TableWriterAlgn* cself, bool commit, uint64_t* rows)
{
    rc_t rc = 0;
    if( cself != NULL ) {
        TableReader_Whack(cself->tmpkey_reader);
        rc = TableWriter_Whack(cself->base, commit, rows);
        free((TableWriterAlgn*)cself);
    }
    return rc;
}

LIB_EXPORT rc_t CC TableWriterAlgn_WriteDefault(const TableWriterAlgn* cself, enum ETableWriterAlgn_ColNames col, const TableWriterData* data)
{
    return TableWriter_ColumnDefault(cself->base, 0, &cself->cols[col], data);
}

LIB_EXPORT rc_t CC TableWriterAlgn_GetNextRowId(const TableWriterAlgn* cself, int64_t* rowid)
{
    return TableWriter_GetNextRowId(cself->base, rowid, cself->cursor_id);
}

LIB_EXPORT rc_t CC TableWriterAlgn_Write(const TableWriterAlgn* cself, const TableWriterAlgnData* data, int64_t* rowid)
{
    rc_t rc = 0;

    if( cself == NULL || data == NULL ) {
        rc = RC( rcAlign, rcType, rcWriting, rcParam, rcNull);
        ALIGN_DBGERR(rc);
    } else if( (rc = TableWriter_OpenRow(cself->base, rowid, cself->cursor_id)) == 0 ) {
        TW_COL_WRITE_VAR(cself->base, cself->cols[ewalgn_cn_PLOIDY], data->ploidy);

        TW_COL_WRITE(cself->base, cself->cols[ewalgn_cn_SEQ_SPOT_ID], data->seq_spot_id);
        TW_COL_WRITE(cself->base, cself->cols[ewalgn_cn_SEQ_READ_ID], data->seq_read_id);
        TW_COL_WRITE(cself->base, cself->cols[ewalgn_cn_READ_START], data->read_start);
        TW_COL_WRITE(cself->base, cself->cols[ewalgn_cn_READ_LEN], data->read_len);
        TW_COL_WRITE(cself->base, cself->cols[ewalgn_cn_TMP_KEY_ID], data->tmp_key_id);

        TW_COL_WRITE(cself->base, cself->cols[ewalgn_cn_HAS_REF_OFFSET], data->has_ref_offset);
        TW_COL_WRITE(cself->base, cself->cols[ewalgn_cn_REF_OFFSET], data->ref_offset);
        TW_COL_WRITE(cself->base, cself->cols[ewalgn_cn_REF_ID], data->ref_id);
        TW_COL_WRITE(cself->base, cself->cols[ewalgn_cn_REF_START], data->ref_start);
        TW_COL_WRITE(cself->base, cself->cols[ewalgn_cn_GLOBAL_REF_START], data->global_ref_start);
        TW_COL_WRITE(cself->base, cself->cols[ewalgn_cn_HAS_MISMATCH], data->has_mismatch);
        TW_COL_WRITE(cself->base, cself->cols[ewalgn_cn_MISMATCH], data->mismatch);
        TW_COL_WRITE(cself->base, cself->cols[ewalgn_cn_REF_LEN], data->ref_len);

        TW_COL_WRITE(cself->base, cself->cols[ewalgn_cn_REF_ORIENTATION], data->ref_orientation);
        TW_COL_WRITE(cself->base, cself->cols[ewalgn_cn_REF_PLOIDY], data->ref_ploidy);
        TW_COL_WRITE(cself->base, cself->cols[ewalgn_cn_MAPQ], data->mapq);

        TW_COL_WRITE(cself->base, cself->cols[ewalgn_cn_MATE_REF_ORIENTATION], data->mate_ref_orientation);
        TW_COL_WRITE(cself->base, cself->cols[ewalgn_cn_MATE_REF_ID], data->mate_ref_id);
        TW_COL_WRITE(cself->base, cself->cols[ewalgn_cn_MATE_REF_POS], data->mate_ref_pos);
        TW_COL_WRITE(cself->base, cself->cols[ewalgn_cn_MATE_ALIGN_ID], data->mate_align_id);
        TW_COL_WRITE(cself->base, cself->cols[ewalgn_cn_TEMPLATE_LEN], data->template_len);

        if( rc == 0 ) {
            rc = TableWriter_CloseRow(cself->base);
        }
    }
    return rc;
}

LIB_EXPORT rc_t CC TableWriterAlgn_TmpKeyStart(const TableWriterAlgn* cself)
{
    rc_t rc = 0;

    if( cself == NULL ) {
        rc = RC(rcAlign, rcType, rcWriting, rcParam, rcNull);
        ALIGN_DBGERR(rc);
    } else if( (cself->options & ewalgn_co_SEQ_SPOT_ID) || !(cself->options & ewalgn_co_TMP_KEY_ID) ) {
        rc = RC(rcAlign, rcType, rcWriting, rcParam, rcViolated);
        ALIGN_DBGERR(rc);
    } else if( (rc = TableWriter_CloseCursor(cself->base, cself->cursor_id, NULL)) == 0 ) {
        TableWriterAlgn* self = (TableWriterAlgn*)cself;
        const VTable* vtbl = NULL;
        memcpy(&self->cols_read_tmpkey, &TableAlgnReadTmpKey_cols, sizeof(TableAlgnReadTmpKey_cols));
        if( (rc = TableWriter_GetVTable(cself->base, &vtbl)) == 0 &&
            (rc = TableReader_Make(&self->tmpkey_reader, vtbl, self->cols_read_tmpkey, 50 * 1024 * 1024)) == 0 ) {
            int64_t v = 0;
            memcpy(self->cols_spotid, &TableWriterAlgn_cols[ewalgn_cn_SEQ_SPOT_ID], sizeof(self->cols_spotid));
            self->cols_spotid[0].flags &= ~ewcol_Ignore;
            rc = TableWriter_AddCursor(self->base, self->cols_spotid,
                                       sizeof(self->cols_spotid) / sizeof(self->cols_spotid[0]),
                                       &self->spotid_cursor_id);
            TW_COL_WRITE_DEF_VAR(self->base, self->spotid_cursor_id, self->cols_spotid[0], v);
        }
    }
    return rc;
}

LIB_EXPORT rc_t CC TableWriterAlgn_TmpKey(const TableWriterAlgn* cself, int64_t rowid, uint32_t* key_id)
{
    rc_t rc = 0;

    if( cself == NULL || rowid == 0 || key_id == NULL ) {
        rc = RC( rcAlign, rcType, rcReading, rcParam, rcNull);
        ALIGN_DBGERR(rc);
    } else if( cself->tmpkey_reader == NULL ) {
        rc = RC( rcAlign, rcType, rcReading, rcMode, rcNotOpen);
        ALIGN_DBGERR(rc);
    } else if( (rc = TableReader_ReadRow(cself->tmpkey_reader, rowid)) == 0 ) {
        memcpy(key_id, cself->cols_read_tmpkey[0].base, sizeof(*key_id));
    }
    return rc;
}

LIB_EXPORT rc_t CC TableWriterAlgn_Write_SpotId(const TableWriterAlgn* cself, int64_t rowid, int64_t spot_id)
{
    rc_t rc = 0;

    if( cself == NULL || rowid == 0 ) {
        rc = RC( rcAlign, rcType, rcWriting, rcParam, rcNull);
        ALIGN_DBGERR(rc);
    } else if( cself->options & ewalgn_co_SEQ_SPOT_ID ) {
        rc = RC( rcAlign, rcType, rcWriting, rcParam, rcViolated);
        ALIGN_DBGERR(rc);
    } else if( (rc = TableWriter_OpenRowId(cself->base, rowid, cself->spotid_cursor_id)) == 0 ) {
        TW_COL_WRITE_VAR(cself->base, cself->cols_spotid[0], spot_id);
        if( rc == 0 ) {
            rc = TableWriter_CloseRow(cself->base);
        }
    }
    return rc;
}

LIB_EXPORT rc_t CC TableWriterAlgn_ReverseCompliment(const INSDC_dna_text* seq, INSDC_dna_text* cmpl, uint32_t len)
{
    rc_t rc = 0;
    uint32_t i;
    static INSDC_dna_text compl[256] = "~";

    if( seq == NULL || compl == NULL ) {
        rc = RC( rcAlign, rcType, rcConverting, rcParam, rcNull);
    } else {
        if( compl[0] == '~' ) {
            INSDC_dna_text* x = (INSDC_dna_text*)&compl[0];
            memset(x, 0, sizeof(compl));
            x['A'] = x['a'] = 'T';
            x['T'] = x['t'] = 'A';
            x['U'] = x['u'] = 'A';
            x['G'] = x['g'] = 'C';
            x['C'] = x['c'] = 'G';
            x['Y'] = x['y'] = 'R';
            x['R'] = x['r'] = 'Y';
            x['S'] = x['s'] = 'S';
            x['W'] = x['w'] = 'W';
            x['K'] = x['k'] = 'M';
            x['M'] = x['m'] = 'K';
            x['B'] = x['b'] = 'V';
            x['D'] = x['d'] = 'H';
            x['H'] = x['h'] = 'D';
            x['V'] = x['v'] = 'B';
            x['N'] = x['n'] = 'N';
        }
        cmpl += len;
        for(i = 0; i < len; i++) {
            *--cmpl= compl[(int)seq[i]];
            if( *cmpl == '\0' ) {
                rc = RC(rcAlign, rcFormatter, rcWriting, rcData, rcInvalid);
                break;
            }
        }
    }
    ALIGN_DBGERR(rc);
    return rc;
}
